/*
    This file is part of SQUIN and it falls under the
    copyright as specified for the whole SQUIN package.
*/
package org.squin.benchmark.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.zip.GZIPOutputStream;

import com.hp.hpl.jena.graph.Factory;
import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.GraphUtil;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;

import org.apache.jena.riot.RDFDataMgr;

import arq.cmdline.ArgDecl;
import arq.cmdline.ArgModuleGeneral;
import arq.cmdline.CmdArgModule;
import arq.cmdline.CmdARQ;
import arq.cmdline.CmdGeneral;
import arq.cmdline.ModQueryIn;


/**
 * ...
 *
 * @author Olaf Hartig (hartig@informatik.hu-berlin.de)
 */
public class PrepareDataset extends CmdARQ
{
	static final protected long seedForRandomizer = 1234L;
	static final protected int numberOfRandomWebsPerDistribution = 2;
	static final protected int numberOfRandomWebsForMainDistribution = 10;

	protected PrepareDatasetModule mod = new PrepareDatasetModule ();

	protected Random random;
	protected Map<String,Graph> graphsBoth100;
	protected Map<String,Graph> graphsBoth000Subj100;
	protected Map<String,Graph> graphsBoth000Subj000;
	protected ArrayList< Map<String,Graph> > graphsBoth066Subj100;
	protected ArrayList< Map<String,Graph> > graphsBoth066Subj066;
	protected ArrayList< Map<String,Graph> > graphsBoth066Subj033;
	protected ArrayList< Map<String,Graph> > graphsBoth066Subj000;
	protected ArrayList< Map<String,Graph> > graphsBoth033Subj100;
	protected ArrayList< Map<String,Graph> > graphsBoth033Subj066;
	protected ArrayList< Map<String,Graph> > graphsBoth033Subj033;
	protected ArrayList< Map<String,Graph> > graphsBoth033Subj000;
	protected ArrayList< Map<String,Graph> > graphsBoth000Subj066;
	protected ArrayList< Map<String,Graph> > graphsBoth000Subj033;
	protected ArrayList< Map<String,Graph> > graphsBoth062Subj047;


	static public void main ( String[] args )
	{
		new PrepareDataset( args ).mainRun();
	}
    
	public PrepareDataset ( String[] args )
	{
		super( args );

		super.addModule( mod );
	}

	@Override
	protected String getCommandName ()
	{
		return getClass().getName();
	}
    
	@Override
	protected String getSummary ()
	{
		return getCommandName() + " --dataset=<dataset file>";
	}

// 	@Override
// 	protected void processModulesAndArgs ()
// 	{
// 		super.processModulesAndArgs();
// 	}
    
	@Override
	protected void exec ()
	{
		System.out.println( "READING INPUT DATASET" );
		Graph inputGraph;
// 		Model inputModel;
		try {
			inputGraph = mod.getDataset();
// 			inputModel = ModelFactory.createModelForGraph( inputGraph );
		}
		catch ( Exception e ) {
			System.err.println( "Creating the input graph/model caused a " + e.getClass().getName() + ": " + e.getMessage() );
			e.printStackTrace( System.err );
			return;
		}

// 		System.out.println( "EXTRACTING RELEVANT URIs" );
// 		Set<Resource> allURIs = extractAllURIs( inputModel );
// for ( Resource uriRes : allURIs ) { System.out.println( "> " + uriRes.getURI() ); }

		System.out.println( "GENERATING OUTPUT" );
		random = new Random( seedForRandomizer );
		initMaps();
		Iterator<Triple> it = GraphUtil.findAll( inputGraph );
		while ( it.hasNext() ) {
			distribute( it.next() );
		}

		System.out.println( "WRITING OUTPUT" );
		writeMaps();

// try {
// 			File outfile = new File( uri.substring(7) );
// 			outfile.getParentFile().mkdirs();
// 			outfile.createNewFile();
// } catch ( Exception e ) { System.err.println( "Creating the output file '" + uri.substring(7) + "' caused a " + e.getClass().getName() + ": " + e.getMessage() ); return; }

	}


	// helpers

	protected Set<Resource> extractAllURIs ( Model inputModel )
	{
		Set<Resource> result = new HashSet<Resource> ();
		Set<Resource> ignored = new HashSet<Resource> ();
		Iterator<Statement> itStmt = inputModel.listStatements();
		while ( itStmt.hasNext() )
		{
			Statement stmt = itStmt.next();

			Resource tmp1 = stmt.getSubject();
			if ( tmp1.isURIResource() && ! result.contains(tmp1) && ! ignored.contains(tmp1) ) {
				if ( ! mustBeIgnored(tmp1.getURI()) ) {
					result.add( tmp1 );
				} else {
					ignored.add( tmp1 );
				}
			}

			RDFNode tmp2 = stmt.getPredicate();
			if ( tmp2.isURIResource() ) {
				tmp1 = tmp2.asResource();
				if ( ! result.contains(tmp1) && ! ignored.contains(tmp1) ) {
					if ( ! mustBeIgnored(tmp1.getURI()) ) {
						result.add( tmp1 );
					} else {
						ignored.add( tmp1 );
					}
				}
			}

			tmp2 = stmt.getObject();
			if ( tmp2.isURIResource() ) {
				tmp1 = tmp2.asResource();
				if ( ! result.contains(tmp1) && ! ignored.contains(tmp1) ) {
					if ( ! mustBeIgnored(tmp1.getURI()) ) {
						result.add( tmp1 );
					} else {
						ignored.add( tmp1 );
					}
				}
			}
		}

		return result;
	}

	protected boolean mustBeIgnored ( String uri )
	{
		// ignore everything that is not an HTTP scheme based URI
		if ( ! uri.startsWith("http://") ) {
			return true;
		}

		// ignore hash URIs and everything with parameters (if not, then creating file names becomes trickier)
		if ( uri.contains("#") || uri.contains("?") ) {
			return true;
		}

		// ignore everything that ends in "/" (and, thus, seems to refer to a Web page)
		if ( uri.endsWith("/") ) {
			return true;
		}

		// if the tool was called without ignore file, we're done
		if ( mod.getIgnoredPrefixes() == null || mod.getIgnoredPrefixes().isEmpty() ) {
			return false;
		}

		for ( String prefix : mod.getIgnoredPrefixes() ) {
			if ( uri.startsWith(prefix) ) {
				return true;
			}
		}

		return false;
	}

	protected void initMaps ()
	{
		graphsBoth100 = new HashMap<String,Graph> ();
		graphsBoth000Subj100 = new HashMap<String,Graph> ();
		graphsBoth000Subj000 = new HashMap<String,Graph> ();

		graphsBoth066Subj100 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth066Subj066 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth066Subj033 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth066Subj000 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth033Subj100 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth033Subj066 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth033Subj033 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth033Subj000 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth000Subj066 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );
		graphsBoth000Subj033 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsPerDistribution );

		for ( int i = 0; i < numberOfRandomWebsPerDistribution; i++ ) {
			graphsBoth066Subj100.add( new HashMap<String,Graph> () );
			graphsBoth066Subj066.add( new HashMap<String,Graph> () );
			graphsBoth066Subj033.add( new HashMap<String,Graph> () );
			graphsBoth066Subj000.add( new HashMap<String,Graph> () );
			graphsBoth033Subj100.add( new HashMap<String,Graph> () );
			graphsBoth033Subj066.add( new HashMap<String,Graph> () );
			graphsBoth033Subj033.add( new HashMap<String,Graph> () );
			graphsBoth033Subj000.add( new HashMap<String,Graph> () );
			graphsBoth000Subj066.add( new HashMap<String,Graph> () );
			graphsBoth000Subj033.add( new HashMap<String,Graph> () );
		}

		graphsBoth066Subj100.trimToSize();
		graphsBoth066Subj066.trimToSize();
		graphsBoth066Subj033.trimToSize();
		graphsBoth066Subj000.trimToSize();
		graphsBoth033Subj100.trimToSize();
		graphsBoth033Subj066.trimToSize();
		graphsBoth033Subj033.trimToSize();
		graphsBoth033Subj000.trimToSize();
		graphsBoth000Subj066.trimToSize();
		graphsBoth000Subj033.trimToSize();

		graphsBoth062Subj047 = new ArrayList<Map<String,Graph>> ( numberOfRandomWebsForMainDistribution );

		for ( int i = 0; i < numberOfRandomWebsForMainDistribution; i++ ) {
			graphsBoth062Subj047.add( new HashMap<String,Graph> () );
		}

		graphsBoth062Subj047.trimToSize();
	}

	protected void distribute ( Triple t )
	{
		String subjURI = ( t.getSubject().isURI() ) ? t.getSubject().getURI() : null;
		String objURI = ( t.getObject().isURI() ) ? t.getObject().getURI() : null;

		add( t, subjURI, graphsBoth100 );
		add( t, objURI,  graphsBoth100 );
		add( t, subjURI, graphsBoth000Subj100 );

		if ( objURI == null ) {  // special case: ensure that each triple goes at least somewhere
			add( t, subjURI, graphsBoth000Subj000 );
		} else {
			add( t, objURI,  graphsBoth000Subj000 );
		}

		for ( int i = 0; i < numberOfRandomWebsPerDistribution; i++ ) {
			maybeAdd( t, subjURI, objURI, graphsBoth066Subj100.get(i), 0.66, 1.0 );
			maybeAdd( t, subjURI, objURI, graphsBoth066Subj066.get(i), 0.66, 0.66 );
			maybeAdd( t, subjURI, objURI, graphsBoth066Subj033.get(i), 0.66, 0.33 );
			maybeAdd( t, subjURI, objURI, graphsBoth066Subj000.get(i), 0.66, 0.0 );
			maybeAdd( t, subjURI, objURI, graphsBoth033Subj100.get(i), 0.33, 1.0 );
			maybeAdd( t, subjURI, objURI, graphsBoth033Subj066.get(i), 0.33, 0.66 );
			maybeAdd( t, subjURI, objURI, graphsBoth033Subj033.get(i), 0.33, 0.33 );
			maybeAdd( t, subjURI, objURI, graphsBoth033Subj000.get(i), 0.33, 0.0 );
			maybeAdd( t, subjURI, objURI, graphsBoth000Subj066.get(i), 0.0, 0.66 );
			maybeAdd( t, subjURI, objURI, graphsBoth000Subj033.get(i), 0.0, 0.33 );
		}

		for ( int i = 0; i < numberOfRandomWebsForMainDistribution; i++ ) {
			maybeAdd( t, subjURI, objURI, graphsBoth062Subj047.get(i), 0.62, 0.47 );
		}
	}

	protected void add ( Triple t, String uri, Map<String,Graph> map )
	{
		if ( uri == null ) {
			return;
		}

		Graph g = map.get( uri );
		if ( g == null ) {
			g = Factory.createGraphMem();
			map.put( uri, g );
		}

		g.add( t );
	}

	protected void maybeAdd ( Triple t, String subjURI, String objURI, Map<String,Graph> map, double bothProb, double subjProb )
	{
		boolean addedToBoth = false;
		if ( bothProb > 0.0 )
		{
			if ( random.nextDouble() < bothProb ) {
				addedToBoth = true;
				add( t, subjURI, map );
				add( t, objURI, map );
			}
		}

		if ( ! addedToBoth )
		{
			if ( objURI == null ) {  // special case: ensure that each triple goes at least somewhere
				add( t, subjURI, map );
			}
			else if ( subjProb >= 1.0 ) {
				add( t, subjURI, map );
			}
			else if ( subjProb <= 0.0 ) {
				add( t, objURI, map );
			}
			else {
				if ( random.nextDouble() < subjProb ) {
					add( t, subjURI, map );
				} else {
					add( t, objURI, map );
				}
			}
		}
	}

	protected void writeMaps ()
	{
		writeMap( graphsBoth100,         new File(mod.getOutDir(),"Both100") );
		writeMap( graphsBoth000Subj100,  new File(mod.getOutDir(),"Both000Subj100") );
		writeMap( graphsBoth000Subj000,  new File(mod.getOutDir(),"Both000Subj000") );

		for ( int i = 0; i < numberOfRandomWebsPerDistribution; i++ ) {
			String ii = String.valueOf( i );
			writeMap( graphsBoth066Subj100.get(i), new File(mod.getOutDir(),"Both066Subj100No"+ii) );
			writeMap( graphsBoth066Subj066.get(i), new File(mod.getOutDir(),"Both066Subj066No"+ii) );
			writeMap( graphsBoth066Subj033.get(i), new File(mod.getOutDir(),"Both066Subj033No"+ii) );
			writeMap( graphsBoth066Subj000.get(i), new File(mod.getOutDir(),"Both066Subj000No"+ii) );
			writeMap( graphsBoth033Subj100.get(i), new File(mod.getOutDir(),"Both033Subj100No"+ii) );
			writeMap( graphsBoth033Subj066.get(i), new File(mod.getOutDir(),"Both033Subj066No"+ii) );
			writeMap( graphsBoth033Subj033.get(i), new File(mod.getOutDir(),"Both033Subj033No"+ii) );
			writeMap( graphsBoth033Subj000.get(i), new File(mod.getOutDir(),"Both033Subj000No"+ii) );
			writeMap( graphsBoth000Subj066.get(i), new File(mod.getOutDir(),"Both000Subj066No"+ii) );
			writeMap( graphsBoth000Subj033.get(i), new File(mod.getOutDir(),"Both000Subj033No"+ii) );
		}

		for ( int i = 0; i < numberOfRandomWebsForMainDistribution; i++ ) {
			String ii = String.valueOf( i );
			writeMap( graphsBoth062Subj047.get(i), new File(mod.getOutDir(),"Both062Subj047No"+ii) );
		}
	}

	protected void writeMap ( Map<String,Graph> map, File dir )
	{
		for ( Map.Entry<String,Graph> entry : map.entrySet() ) {
			if ( ! mustBeIgnored(entry.getKey()) ) {
				File f = createOutputFile( dir, entry.getKey() );
				if ( f != null ) {
					GZIPOutputStream s = null;
					try {
						s = new GZIPOutputStream( new FileOutputStream(f) );
						ModelFactory.createModelForGraph( entry.getValue() ).write( s, "TTL" );
						s.close();
					}
					catch ( Exception e ) {
						System.err.println( "Writing to output file '" + f.getAbsolutePath() + "' caused a " + e.getClass().getName() + ": " + e.getMessage() );
						if ( s != null ) {
							try { s.close(); } catch ( Exception e2 ) { /* ignore */ }
						}
					}
				}
			}
			else {
				System.out.println( "Ignore: " + entry.getKey() );
			}
		}
	}

	protected File createOutputFile ( File dir, String uri )
	{
		String relativeFilename = uri.substring(7);  // cut off the "http://" part
		File f = new File( dir, relativeFilename );
		try {
			f.getParentFile().mkdirs();
		}
		catch ( Exception e ) {
			System.err.println( "Creating the directories for output file '" + f.getAbsolutePath() + "' caused a " + e.getClass().getName() + ": " + e.getMessage() );
			return null;
		}

		try {
			f.createNewFile();
		}
		catch ( Exception e ) {
			System.err.println( "Creating the output file '" + f.getAbsolutePath() + "' caused a " + e.getClass().getName() + ": " + e.getMessage() );
			return null;
		}

		return f;
	}


public class PrepareDatasetModule implements ArgModuleGeneral
{
	final protected ArgDecl datasetDecl = new ArgDecl( ArgDecl.HasValue, "dataset" );
	final protected ArgDecl outdirDecl = new ArgDecl( ArgDecl.HasValue, "outdir" );
	final protected ArgDecl ignoreDecl = new ArgDecl( ArgDecl.HasValue, "ignore" );

	private String datasetFileName;
	private Graph dataset;
	private File outdir;
	private Set<String> ignoreList;


	// implementation of the ArgModuleGeneral interface

	public void registerWith ( CmdGeneral cmdLine )
	{
		cmdLine.getUsage().startCategory( "Prepare dataset" );
		cmdLine.add( datasetDecl,
		             "--dataset",
		             "RDF document containing the benchmark dataset as generated by BSBM (mandatory)") ;
		cmdLine.add( outdirDecl,
		             "--outdir",
		             "Main directory for the output (mandatory)") ;
		cmdLine.add( ignoreDecl,
		             "--ignore",
		             "A file in which each line contains a prefix for URIs that have to be ignored; i.e., for which no document has to be generated (optional)") ;
	}


	// implementation of the ArgModule interface

	public void processArgs ( CmdArgModule cmdLine )
	{
		dataset = null;
		outdir = null;

		if ( cmdLine.contains(datasetDecl) )
		{
			datasetFileName = cmdLine.getValue( datasetDecl );
			File datasetFile = new File( datasetFileName );
			if ( ! datasetFile.exists() || ! datasetFile.canRead() ) {
				cmdLine.cmdError( "The given dataset file (" + datasetFile.toString() + ") does not exist or is not readable" );
			}
		}
		else {
			cmdLine.cmdError( "No dataset specified" );
		}

		if ( cmdLine.contains(outdirDecl) )
		{
			String outdirName = cmdLine.getValue( outdirDecl );
			outdir = new File( outdirName );
		}
		else {
			cmdLine.cmdError( "No output directory specified" );
		}

		if ( cmdLine.contains(ignoreDecl) )
		{
			String ignoreFileName = cmdLine.getValue( ignoreDecl );
			File ignoreFile = new File( ignoreFileName );
			if ( ! ignoreFile.exists() || ! ignoreFile.canRead() ) {
				cmdLine.cmdError( "The given ignore file (" + ignoreFile.toString() + ") does not exist or is unreadable" );
			}

			ignoreList = new HashSet<String> ();
			try {
				BufferedReader r = new BufferedReader( new FileReader(ignoreFile) );
				String line;
				while ( (line=r.readLine()) != null ) {
					ignoreList.add( line );
				}
				r.close();
			}
			catch ( IOException e ) {
				cmdLine.cmdError( "Reading the ignore file (" + ignoreFile.toString() + ") caused a " + e.getClass().getName() + ": " + e.getMessage() );
			}
		}
		else {
			ignoreList = null;
		}
	}


	// accessors

	public Graph getDataset ()
	{
		if ( dataset == null ) {
			dataset = RDFDataMgr.loadGraph( datasetFileName );
		}

		return dataset;
	}

	public File getOutDir ()
	{
		return outdir;
	}

	public Set<String> getIgnoredPrefixes ()
	{
		return ignoreList;
	}
}


}
